#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include "Random.h"

//-------------------------------------------------------------------------------------------

Random *Random::m_instance = 0;

//-------------------------------------------------------------------------------------------

Random::Random()
{
	reset();
}

//-------------------------------------------------------------------------------------------

Random::~Random()
{}

//-------------------------------------------------------------------------------------------

Random *Random::instance()
{
	if(m_instance == 0)
	{
		m_instance = new Random;
	}
	return m_instance;
}

//-------------------------------------------------------------------------------------------

void Random::reset()
{
	m_mti = RANDOM_MT19937_LENGTH_N + 1;
	initializePRNG(5489UL);
}

//-------------------------------------------------------------------------------------------

void Random::seed(int32_t seedValue)
{
	initializePRNG(static_cast<uint32_t>(seedValue));
}

//-------------------------------------------------------------------------------------------

void Random::seed(uint32_t seedValue)
{
	initializePRNG(seedValue);
}

//-------------------------------------------------------------------------------------------

void Random::seed(const int32_t *key,int keyLength)
{
	initializePRNGArray(reinterpret_cast<const uint32_t *>(key),keyLength);
}

//-------------------------------------------------------------------------------------------

void Random::seed(const uint32_t *key,int keyLength)
{
	initializePRNGArray(key,keyLength);
}

//-------------------------------------------------------------------------------------------

void Random::initializePRNG(uint32_t s)
{
	m_mt[0] = s & 0xffffffffUL;
	for(m_mti=1;m_mti < RANDOM_MT19937_LENGTH_N;m_mti++)
	{
		m_mt[m_mti] = (1812433253UL * (m_mt[m_mti-1] ^ (m_mt[m_mti-1] >> 30)) + m_mti);
		m_mt[m_mti] &= 0xffffffffUL;
	}
}

//-------------------------------------------------------------------------------------------

void Random::initializePRNGArray(const uint32_t *key,int keyLength)
{
	int i = 1;
	int j = 0;
	int k = (RANDOM_MT19937_LENGTH_N > keyLength) ? RANDOM_MT19937_LENGTH_N : keyLength;
	
	initializePRNG(19650218UL);
	
	while(k)
	{
		m_mt[i] = (m_mt[i] ^ ((m_mt[i-1] ^ (m_mt[i-1] >> 30)) * 1664525UL)) + key[j] + j;
		m_mt[i] &= 0xffffffffUL;
		i++;
		j++;
		if(i>=RANDOM_MT19937_LENGTH_N)
		{
			m_mt[0] = m_mt[RANDOM_MT19937_LENGTH_N-1];
			i=1; 
		}
        if(j>=keyLength)
        {
        	j=0;
        }
		k--;	
	}
	
	for(k=RANDOM_MT19937_LENGTH_N-1;k;k--)
	{
		m_mt[i] = (m_mt[i] ^ ((m_mt[i-1] ^ (m_mt[i-1] >> 30)) * 1566083941UL)) - i;
		m_mt[i] &= 0xffffffffUL;
		i++;
		if(i>=RANDOM_MT19937_LENGTH_N)
		{
			m_mt[0] = m_mt[RANDOM_MT19937_LENGTH_N-1];
			i=1;
		}
	}
	
	m_mt[0] = 0x80000000UL;
}
//-------------------------------------------------------------------------------------------

uint32_t Random::generateRandomUInt32()
{
	static uint32_t mag01[2] = {0x00000000UL, RANDOM_MT19937_MATRIX_A};

	uint32_t y;

	if(m_mti>=RANDOM_MT19937_LENGTH_N)
	{
		int kk;

		if(m_mti==RANDOM_MT19937_LENGTH_N+1)
		{
			initializePRNG(5489UL);
		}

		for(kk=0;kk<RANDOM_MT19937_LENGTH_N-RANDOM_MT19937_LENGTH_M;kk++)
		{
			y = (m_mt[kk] & RANDOM_MT19937_UPPER_MASK) | (m_mt[kk+1] & RANDOM_MT19937_LOWER_MASK);
			m_mt[kk] = m_mt[kk+RANDOM_MT19937_LENGTH_M] ^ (y >> 1) ^ mag01[y & 0x1UL];
		}
		
		for(;kk<RANDOM_MT19937_LENGTH_N-1;kk++)
		{
			y = (m_mt[kk] & RANDOM_MT19937_UPPER_MASK) | (m_mt[kk+1] & RANDOM_MT19937_LOWER_MASK);
			m_mt[kk] = m_mt[kk+(RANDOM_MT19937_LENGTH_M-RANDOM_MT19937_LENGTH_N)] ^ (y >> 1) ^ mag01[y & 0x1UL];
		}
		
		y = (m_mt[RANDOM_MT19937_LENGTH_N-1] & RANDOM_MT19937_UPPER_MASK) | (m_mt[0] & RANDOM_MT19937_LOWER_MASK);
		m_mt[RANDOM_MT19937_LENGTH_N-1] = m_mt[RANDOM_MT19937_LENGTH_M-1] ^ (y >> 1) ^ mag01[y & 0x1UL];

		m_mti = 0;
	}

	y = m_mt[m_mti++];

	y ^= (y >> 11);
	y ^= (y << 7) & 0x9d2c5680UL;
	y ^= (y << 15) & 0xefc60000UL;
	y ^= (y >> 18);

	return y;
}

//-------------------------------------------------------------------------------------------
// generates a random number on [0,0xffffffff]-interval
//-------------------------------------------------------------------------------------------

uint32_t Random::randomUInt32()
{
	return generateRandomUInt32();
}

//-------------------------------------------------------------------------------------------
// generates a random number on [0,0x7fffffff]-interval
//-------------------------------------------------------------------------------------------

int32_t Random::randomInt31()
{
	return static_cast<int32_t>(generateRandomUInt32() >> 1);
}

//-------------------------------------------------------------------------------------------

double Random::randomReal1()
{
	return static_cast<double>(generateRandomUInt32()) * (1.0/4294967295.0);
}

//-------------------------------------------------------------------------------------------
